import logging
import os

import jsonschema

from octue.compatibility import warn_if_incompatible


VALID_EVENT_KINDS = {"delivery_acknowledgement", "heartbeat", "log_record", "monitor_message", "exception", "result"}

SERVICE_COMMUNICATION_SCHEMA = {"$ref": "https://jsonschema.registry.octue.com/octue/service-communication/0.10.0.json"}
SERVICE_COMMUNICATION_SCHEMA_INFO_URL = "https://strands.octue.com/octue/service-communication"
SERVICE_COMMUNICATION_SCHEMA_VERSION = os.path.splitext(SERVICE_COMMUNICATION_SCHEMA["$ref"])[0].split("/")[-1]

# Instantiate a JSON schema validator to cache the service communication schema. This avoids downloading it from the
# registry every time a message is validated against it.
jsonschema.Draft202012Validator.check_schema(SERVICE_COMMUNICATION_SCHEMA)
jsonschema_validator = jsonschema.Draft202012Validator(SERVICE_COMMUNICATION_SCHEMA)

logger = logging.getLogger(__name__)


def is_event_valid(event, attributes, recipient, parent_sdk_version, child_sdk_version, schema=None):
    """Check if the event and its attributes are valid according to the Octue services communication schema.

    :param dict event: the event to validate
    :param dict attributes: the attributes of the event to validate
    :param octue.cloud.pub_sub.service.Service recipient: the service receiving and validating the event
    :param str parent_sdk_version: the semantic version of Octue SDK running on the parent
    :param str child_sdk_version: the semantic version of Octue SDK running on the child
    :param dict|None schema: the schema to validate the event and its attributes against; if `None`, this defaults to the service communication schema used in this version of Octue SDK
    :return bool: `True` if the event and its attributes are valid
    """
    try:
        raise_if_event_is_invalid(
            event,
            attributes,
            recipient,
            parent_sdk_version,
            child_sdk_version,
            schema=schema,
        )
    except jsonschema.ValidationError:
        return False

    return True


def raise_if_event_is_invalid(event, attributes, recipient, parent_sdk_version, child_sdk_version, schema=None):
    """Raise an error if the event or its attributes aren't valid according to the Octue services communication schema.

    :param dict event: the event to validate
    :param dict attributes: the attributes of the event to validate
    :param octue.cloud.pub_sub.service.Service recipient: the service receiving and validating the event
    :param str parent_sdk_version: the semantic version of Octue SDK running on the parent
    :param str child_sdk_version: the semantic version of Octue SDK running on the child
    :param dict|None schema: the schema to validate the event and its attributes against; if `None`, this defaults to the service communication schema used in this version of Octue SDK
    :raise jsonschema.ValidationError: if the event or its attributes are invalid
    :return None:
    """
    # Transform attributes to a dictionary in the case they're a different kind of mapping.
    data = {"event": event, "attributes": dict(attributes)}

    if schema is None:
        schema = SERVICE_COMMUNICATION_SCHEMA

    try:
        # If the schema is the official service communication schema, use the cached validator.
        if schema == SERVICE_COMMUNICATION_SCHEMA:
            jsonschema_validator.validate(data)

        # Otherwise, use uncached validation.
        else:
            jsonschema.validate(data, schema)

    except jsonschema.ValidationError as error:
        warn_if_incompatible(parent_sdk_version=parent_sdk_version, child_sdk_version=child_sdk_version)

        logger.exception(
            "%r received an event that doesn't conform with version %s of the service communication schema (%s): %r.",
            recipient,
            SERVICE_COMMUNICATION_SCHEMA_VERSION,
            SERVICE_COMMUNICATION_SCHEMA_INFO_URL,
            event,
        )

        raise error



class EventCounter:
    """A mutable counter for keeping track of the emission order of events. This is used in the `Service` class instead
    of an integer because it is mutable and can be passed to the `Service._emit_event` method and incremented as
    events are emitted.

    :return None:
    """

    def __init__(self):
        self.count = 0

    def __iadd__(self, other):
        """Increment the counter by an integer.

        :return octue.cloud.events.counter.EventCounter: the event counter with its count updated
        """
        if not isinstance(other, int):
            raise ValueError(f"Event counters can only be incremented by an integer; received {other!r}.")

        self.count += other
        return self

    def __int__(self):
        """Get the counter as an integer.

        :return int: the counter as an integer
        """
        return int(self.count)

    def __repr__(self):
        """Represent the counter as a string.

        :return str: the counter represented as a string.
        """
        return f"<{type(self).__name__}(count={self.count})"



import base64
import json
import logging
import time
from datetime import datetime, timedelta

from google.api_core import retry
from google.cloud.pubsub_v1 import SubscriberClient

from octue.cloud.events.handler import AbstractEventHandler
from octue.cloud.events.validation import SERVICE_COMMUNICATION_SCHEMA
from octue.utils.decoders import OctueJSONDecoder
from octue.utils.objects import getattr_or_subscribe
from octue.utils.threads import RepeatingTimer


logger = logging.getLogger(__name__)

MAX_SIMULTANEOUS_MESSAGES_PULL = 50


def extract_event_and_attributes_from_pub_sub_message(message):
    """Extract an Octue service event and its attributes from a Google Pub/Sub message in either direct Pub/Sub format
    or in the Google Cloud Run format.

    :param dict|google.cloud.pubsub_v1.subscriber.message.Message message: the message in Google Cloud Run format or Google Pub/Sub format
    :return (any, dict): the extracted event and its attributes
    """
    # Cast attributes to a dictionary to avoid defaultdict-like behaviour from Pub/Sub message attributes container.
    attributes = dict(getattr_or_subscribe(message, "attributes"))

    # Deserialise the `order` and `forward_logs` fields if they're present (don't assume they are before validation).
    if attributes.get("order"):
        attributes["order"] = int(attributes["order"])

    # Required for question events.
    if attributes.get("sender_type") == "PARENT":
        forward_logs = attributes.get("forward_logs")

        if forward_logs:
            attributes["forward_logs"] = bool(int(forward_logs))
        else:
            attributes["forward_logs"] = None

    try:
        # Parse event directly from Pub/Sub or Dataflow.
        event = json.loads(message.data.decode(), cls=OctueJSONDecoder)
    except Exception:
        # Parse event from Google Cloud Run.
        event = json.loads(base64.b64decode(message["data"]).decode("utf-8").strip(), cls=OctueJSONDecoder)

    return event, attributes


class GoogleCloudPubSubEventHandler(AbstractEventHandler):
    """A synchronous handler for events received as Google Pub/Sub messages from a pull subscription.

    :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription messages are pulled from
    :param octue.cloud.pub_sub.service.Service recipient: the `Service` instance that's receiving the events
    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param bool record_events: if `True`, record received events in the `received_events` attribute
    :param dict|None event_handlers: a mapping of event type names to callables that handle each type of event. The handlers must not mutate the events.
    :param dict|str schema: the JSON schema to validate events against
    :param int|float skip_missing_events_after: the number of seconds after which to skip any events if they haven't arrived but subsequent events have
    :return None:
    """

    def __init__(
        self,
        subscription,
        recipient,
        handle_monitor_message=None,
        record_events=True,
        event_handlers=None,
        schema=SERVICE_COMMUNICATION_SCHEMA,
        skip_missing_events_after=10,
    ):
        self.subscription = subscription

        super().__init__(
            recipient,
            handle_monitor_message=handle_monitor_message,
            record_events=record_events,
            event_handlers=event_handlers,
            schema=schema,
            skip_missing_events_after=skip_missing_events_after,
        )

        self._subscriber = SubscriberClient()
        self._heartbeat_checker = None
        self._last_heartbeat = None
        self._alive = True

    @property
    def total_run_time(self):
        """The amount of time elapsed since `self.handle_events` was called. If it hasn't been called yet, this is
        `None`.

        :return float|None: the amount of time [s] since `self.handle_events` was called
        """
        if self._start_time is None:
            return None

        return time.perf_counter() - self._start_time

    @property
    def _time_since_last_heartbeat(self):
        """The amount of time since the last heartbeat was received. If no heartbeat has been received, this is `None`.

        :return datetime.timedelta|None:
        """
        if not self._last_heartbeat:
            return None

        return datetime.now() - self._last_heartbeat

    def handle_events(self, timeout=60, maximum_heartbeat_interval=300):
        """Pull events from the subscription and handle them in the order they were sent until a "result" event is
        handled, then return the handled result.

        :param float|None timeout: how long to wait for an answer before raising a `TimeoutError`
        :param int|float maximum_heartbeat_interval: the maximum amount of time [s] allowed between child heartbeats before an error is raised
        :raise TimeoutError: if the timeout is exceeded before receiving the final event
        :return dict: the handled "result" event
        """
        super().handle_events()

        self._heartbeat_checker = RepeatingTimer(
            interval=maximum_heartbeat_interval,
            function=self._monitor_heartbeat,
            kwargs={"maximum_heartbeat_interval": maximum_heartbeat_interval},
        )

        try:
            self._heartbeat_checker.daemon = True
            self._heartbeat_checker.start()

            while self._alive:
                pull_timeout = self._check_timeout_and_get_pull_timeout(timeout)
                self._pull_and_enqueue_available_events(timeout=pull_timeout)
                result = self._attempt_to_handle_waiting_events()

                if result is not None:
                    return result

        finally:
            self._heartbeat_checker.cancel()
            self._subscriber.close()

        raise TimeoutError(
            f"No heartbeat has been received within the maximum allowed interval of {maximum_heartbeat_interval}s."
        )

    def _monitor_heartbeat(self, maximum_heartbeat_interval):
        """Change the alive status to `False` and cancel the heartbeat checker if a heartbeat hasn't been received
        within the maximum allowed time interval since the last received heartbeat.

        :param float|int maximum_heartbeat_interval: the maximum amount of time [s] allowed between child heartbeats without raising an error
        :return None:
        """
        maximum_heartbeat_interval = timedelta(seconds=maximum_heartbeat_interval)

        if self._last_heartbeat and self._time_since_last_heartbeat <= maximum_heartbeat_interval:
            self._alive = True
            return

        self._alive = False
        self._heartbeat_checker.cancel()

    def _check_timeout_and_get_pull_timeout(self, timeout):
        """Check if the message handling timeout has been exceeded and, if it hasn't, calculate and return the timeout
        for the next message pull. If the timeout has been exceeded, raise an error.

        :param int|float|None timeout: the timeout [s] for handling all messages, or `None` if there's no timeout
        :raise TimeoutError: if the timeout has been exceeded
        :return int|float|None: the timeout for the next message pull [s], or `None` if there's no timeout
        """
        if timeout is None:
            return None

        # Get the total run time once in case it's very close to the timeout - this rules out a negative pull timeout
        # being returned below.
        total_run_time = self.total_run_time

        if total_run_time > timeout:
            raise TimeoutError(f"No final result received from {self.subscription.topic!r} after {timeout} seconds.")

        return timeout - total_run_time

    def _pull_and_enqueue_available_events(self, timeout):
        """Pull as many events from the subscription as are available and enqueue them in `self.waiting_events`,
        raising a `TimeoutError` if the timeout is exceeded before succeeding.

        :param float|None timeout: how long to wait for the event [s] before raising a `TimeoutError`
        :raise TimeoutError|concurrent.futures.TimeoutError: if the timeout is exceeded
        :return None:
        """
        pull_start_time = time.perf_counter()
        attempt = 1

        while self._alive:
            logger.debug("Pulling events from Google Pub/Sub: attempt %d.", attempt)

            pull_response = self._subscriber.pull(
                request={"subscription": self.subscription.path, "max_messages": MAX_SIMULTANEOUS_MESSAGES_PULL},
                retry=retry.Retry(),
            )

            if len(pull_response.received_messages) > 0:
                break
            else:
                logger.debug("Google Pub/Sub pull response timed out early.")
                attempt += 1

                pull_run_time = time.perf_counter() - pull_start_time

                if timeout is not None and pull_run_time > timeout:
                    raise TimeoutError(f"No message received from {self.subscription.topic!r} after {timeout} seconds.")

        if not pull_response.received_messages:
            return

        self._subscriber.acknowledge(
            request={
                "subscription": self.subscription.path,
                "ack_ids": [message.ack_id for message in pull_response.received_messages],
            }
        )

        for event in pull_response.received_messages:
            self._extract_and_enqueue_event(event)

    def _extract_event_and_attributes(self, container):
        """Extract an event and its attributes from a Pub/Sub message.

        :param dict container: a Pub/Sub message
        :return (any, dict): the event and its attributes
        """
        return extract_event_and_attributes_from_pub_sub_message(container.message)



import logging
import re


ANSI_ESCAPE_SEQUENCES_PATTERN = r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])"


class GoogleCloudPubSubHandler(logging.Handler):
    """A log handler that publishes log records to a Google Cloud Pub/Sub topic.

    :param callable event_emitter: the `_emit_event` method of the service that instantiated this instance
    :param str question_uuid: the UUID of the question to handle log records for
    :param str originator: the SRUID of the service that asked the question these log records are related to
    :param str recipient: the SRUID of the service to send these log records to
    :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
    :param float timeout: timeout in seconds for attempting to publish each log record
    :return None:
    """

    def __init__(self, event_emitter, question_uuid, originator, recipient, order, timeout=60, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.question_uuid = question_uuid
        self.originator = originator
        self.recipient = recipient
        self.order = order
        self.timeout = timeout
        self._emit_event = event_emitter

    def emit(self, record):
        """Serialise the log record as a dictionary and publish it to the topic.

        :param logging.LogRecord record:
        :return None:
        """
        try:
            self._emit_event(
                {
                    "kind": "log_record",
                    "log_record": self._convert_log_record_to_primitives(record),
                },
                originator=self.originator,
                recipient=self.recipient,
                order=self.order,
                attributes={
                    "question_uuid": self.question_uuid,
                    "sender_type": "CHILD",  # The sender type is repeated here as a string to avoid a circular import.
                },
            )

        except Exception:  # noqa
            self.handleError(record)

    def _convert_log_record_to_primitives(self, log_record):
        """Convert a log record to JSON-serialisable primitives by interpolating the args into the message, and
        removing the exception info, which is potentially not JSON-serialisable. This is similar to the approach in
        `logging.handlers.SocketHandler.makePickle`. Also strip any ANSI escape sequences from the message.

        :param logging.LogRecord log_record:
        :return dict:
        """
        serialised_record = vars(log_record)

        serialised_record["msg"] = re.compile(ANSI_ESCAPE_SEQUENCES_PATTERN).sub("", log_record.getMessage())
        serialised_record["args"] = None
        serialised_record["exc_info"] = None
        serialised_record.pop("message", None)

        if not serialised_record.get("levelno"):
            serialised_record["levelno"] = logging.INFO

        return serialised_record



import logging

from octue.cloud.pub_sub.service import Service
from octue.cloud.service_id import create_sruid, get_sruid_parts
from octue.configuration import load_service_and_app_configuration
from octue.resources.service_backends import GCPPubSubBackend
from octue.runner import Runner
from octue.utils.objects import get_nested_attribute


logger = logging.getLogger(__name__)


DEFAULT_SERVICE_CONFIGURATION_PATH = "octue.yaml"


def answer_question(question, project_name):
    """Answer a question sent to an app deployed in Google Cloud.

    :param dict|tuple question:
    :param str project_name:
    :return None:
    """
    service_configuration, app_configuration = load_service_and_app_configuration(DEFAULT_SERVICE_CONFIGURATION_PATH)
    service_namespace, service_name, service_revision_tag = get_sruid_parts(service_configuration)

    service_sruid = create_sruid(
        namespace=service_namespace,
        name=service_name,
        revision_tag=service_revision_tag,
    )

    service = Service(service_id=service_sruid, backend=GCPPubSubBackend(project_name=project_name))
    question_uuid = get_nested_attribute(question, "attributes.question_uuid")

    try:
        runner = Runner.from_configuration(
            service_configuration=service_configuration,
            app_configuration=app_configuration,
            project_name=project_name,
            service_id=service_sruid,
        )

        service.run_function = runner.run
        service.answer(question)
        logger.info("Analysis successfully run and response sent for question %r.", question_uuid)

    except BaseException as error:  # noqa
        service.send_exception(question_uuid=question_uuid, originator="UNKNOWN")
        logger.exception(error)



from .base import MixinBase
from .cloud_pathable import CloudPathable
from .cool_nameable import CoolNameable
from .filterable import Filterable
from .hashable import Hashable
from .identifiable import Identifiable
from .labelable import Labelable
from .metadata import Metadata
from .serialisable import Serialisable
from .taggable import Taggable


__all__ = (
    "CloudPathable",
    "CoolNameable",
    "Filterable",
    "Hashable",
    "Identifiable",
    "Labelable",
    "Metadata",
    "MixinBase",
    "Serialisable",
    "Taggable",
)



import logging
import logging.handlers
import os
from urllib.parse import urlparse

from octue.definitions import GOOGLE_COMPUTE_PROVIDERS


if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") in GOOGLE_COMPUTE_PROVIDERS:
    # Google Cloud logs don't support colour currently - provide a no-operation function.
    colourise = lambda string, text_colour=None, background_colour=None: string
else:
    from octue.utils.colour import colourise  # noqa


# Logging format for analysis runs. All handlers should use this logging format to make logs consistently parseable.
LOG_RECORD_ATTRIBUTES_WITH_TIMESTAMP = ["%(asctime)s", "%(levelname)s", "%(name)s"]
LOG_RECORD_ATTRIBUTES_WITHOUT_TIMESTAMP = LOG_RECORD_ATTRIBUTES_WITH_TIMESTAMP[1:]

# "colorblind" colour palette from seaborn/matplotlib.
COLOUR_PALETTE = ["0173b2", "de8f05", "029e73", "d55e00", "cc78bc", "ca9161", "fbafe4", "949494", "ece133", "56b4e9"]


def should_use_octue_log_handler():
    """Return `True` if the `USE_OCTUE_LOG_HANDLER` environment variable is set to "1" or is unspecified; return `False`
    if it's set to "0" or anything else.

    :return bool: `True` if the `USE_OCTUE_LOG_HANDLER` environment variable is set to "1" or is unspecified
    """
    return int(os.environ.get("USE_OCTUE_LOG_HANDLER", "1")) == 1


def create_octue_formatter(
    *log_record_attributes,
    include_line_number=False,
    include_process_name=False,
    include_thread_name=False,
    use_colour=True,
):
    """Create a log formatter from the given log record attributes that delimits the attributes with space-padded pipes
    and encapsulates the whole log message context in square brackets before adding the message at the end. e.g. if the
    attributes are `["%(asctime)s", "%(levelname)s", "%(name)s"]`, the formatter would format log messages as e.g.
    `[2021-06-29 11:58:10,985 | INFO | octue.runner] This is a log message.`

    :param log_record_attributes: any number of iterables of log record attribute names to use as context for every log message that the formatter is applied to; each iterable is interpreted as a different section by the formatter
    :param bool include_line_number: if `True`, include the line number in the log context
    :param bool include_process_name: if `True`, include the process name in the log context
    :param bool include_thread_name: if `True`, include the thread name in the log context
    :param bool use_colour: if `True`, use ANSI colour codes to colour the logs
    :return logging.Formatter:
    """
    global colourise

    extra_attributes = []

    if include_line_number:
        extra_attributes.append("%(lineno)d")
    if include_process_name:
        extra_attributes.append("%(processName)s")
    if include_thread_name:
        extra_attributes.append("%(threadName)s")

    if not use_colour:
        colourise = lambda string, text_colour=None, background_colour=None: string

    if len(log_record_attributes) > 1:
        extra_sections = [
            " ".join(
                colourise(
                    "[" + " | ".join(attributes_section) + "]",
                    text_colour=COLOUR_PALETTE[2],
                )
                for attributes_section in log_record_attributes[1:]
            )
        ]

    else:
        extra_sections = []

    return logging.Formatter(
        " ".join(
            [
                colourise(
                    "[" + " | ".join(log_record_attributes[0] + extra_attributes) + "]",
                    text_colour=COLOUR_PALETTE[0],
                ),
                *extra_sections,
                "%(message)s",
            ]
        )
    )


def apply_log_handler(
    logger_name=None,
    logger=None,
    handler=None,
    log_level=logging.INFO,
    formatter=None,
    include_line_number=False,
    include_process_name=False,
    include_thread_name=False,
):
    """Apply a log handler with the given formatter to the logger with the given name. By default, the default Octue log
    handler is used on the root logger.

    :param str|None logger_name: the name of the logger to apply the handler to; if this and `logger` are `None`, the root logger is used
    :param logging.Logger|None logger: the logger instance to apply the handler to (takes precedence over a logger name)
    :param logging.Handler|None handler: The handler to use. If `None`, the default `StreamHandler` will be attached.
    :param int|str log_level: ignore log messages below this level
    :param logging.Formatter|None formatter: if provided, this formatter is used and the other formatting options are ignored
    :param bool include_line_number: if `True`, include the line number in the log context
    :param bool include_process_name: if `True`, include the process name in the log context
    :param bool include_thread_name: if `True`, include the thread name in the log context
    :return logging.Handler:
    """
    logger = logger or logging.getLogger(name=logger_name)
    handler = handler or logging.StreamHandler()

    for existing_handler in logger.handlers:
        if type(existing_handler).__name__ == "StreamHandler" and type(handler).__name__ == "StreamHandler":
            logger.removeHandler(existing_handler)

    if formatter is None:
        formatter = create_octue_formatter(
            get_log_record_attributes_for_environment(),
            include_line_number=include_line_number,
            include_process_name=include_process_name,
            include_thread_name=include_thread_name,
        )

    handler.setFormatter(formatter)
    handler.setLevel(log_level)

    logger.addHandler(handler)
    logger.setLevel(log_level)

    for handler in logger.handlers:
        if type(handler).__name__ == "SocketHandler":
            # Log locally that a remote logger will be used.
            local_logger = logging.getLogger(__name__)
            temporary_handler = logging.StreamHandler()
            temporary_handler.setFormatter(formatter)
            temporary_handler.setLevel(log_level)
            local_logger.addHandler(temporary_handler)
            local_logger.setLevel(log_level)
            local_logger.info("Logs streaming to %s:%s", handler.host, str(handler.port))
            local_logger.removeHandler(temporary_handler)
            break

    return handler


def get_remote_handler(
    logger_uri,
    formatter=None,
    include_line_number=False,
    include_process_name=False,
    include_thread_name=False,
):
    """Get a log handler for streaming logs to a remote URI accessed via HTTP or HTTPS. The default octue log formatter
    is used if no formatter is provided.

    :param str logger_uri: the URI to stream the logs to
    :param logging.Formatter|None formatter: if provided, this formatter is used and the other formatting options are ignored
    :param bool include_line_number: if `True`, include the line number in the log context
    :param bool include_process_name: if `True`, include the process name in the log context
    :param bool include_thread_name: if `True`, include the thread name in the log context
    :return logging.Handler:
    """
    parsed_uri = urlparse(logger_uri)

    if parsed_uri.scheme not in {"ws", "wss"}:
        raise ValueError(
            f"Only WS and WSS protocols currently supported for remote logger URI. Received {logger_uri!r}."
        )

    handler = logging.handlers.SocketHandler(host=parsed_uri.hostname, port=parsed_uri.port)

    formatter = formatter or create_octue_formatter(
        get_log_record_attributes_for_environment(),
        include_line_number=include_line_number,
        include_process_name=include_process_name,
        include_thread_name=include_thread_name,
    )

    handler.setFormatter(formatter)
    return handler


def get_log_record_attributes_for_environment():
    """Get the correct log record attributes for the environment. If the environment is Google Cloud Run, get log record
    attributes not including the timestamp in the log context to avoid the date appearing twice in the Google Cloud Run
    logs (Google adds its own timestamp to log messages). Otherwise, get log record attributes including the timestamp.

    :return list:
    """
    if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") in GOOGLE_COMPUTE_PROVIDERS:
        return LOG_RECORD_ATTRIBUTES_WITHOUT_TIMESTAMP

    return LOG_RECORD_ATTRIBUTES_WITH_TIMESTAMP


class AnalysisLogFormatterSwitcher:
    """A context manager that, when activated, removes any formatters from the logger's handlers, adds any extra
    handlers provided, and adds formatters that include the analysis ID to all the handlers. On leaving the context, the
    logger is restored to its initial state.

    If the `USE_OCTUE_LOG_HANDLER` environment variable is not set to "1" and the logger has handlers, their formatters
    are left unchanged and the extra log handlers are added with the logger's zeroth handler's formatter applied to
    them.

    :param str analysis_id: the ID of the analysis to add to the log formatters
    :param logger.Logger logger: the logger whose handlers' formatters should be switched
    :param str|int analysis_log_level: the log level to apply to any extra log handlers
    :param list(logging.Handler) extra_log_handlers: any extra log handlers to add to the logger
    :return None:
    """

    def __init__(self, analysis_id, logger, analysis_log_level, extra_log_handlers=None):
        self.analysis_id = analysis_id
        self.logger = logger
        self.analysis_log_level = analysis_log_level
        self.extra_log_handlers = extra_log_handlers or []
        self.initial_formatters = [(handler, handler.formatter) for handler in self.logger.handlers]

        if should_use_octue_log_handler() or not self.logger.handlers:
            # Create formatters that include the analysis ID in the logging metadata.
            self.coloured_analysis_formatter = create_octue_formatter(
                get_log_record_attributes_for_environment(),
                [f"analysis-{self.analysis_id}"],
            )

            self.uncoloured_analysis_formatter = create_octue_formatter(
                get_log_record_attributes_for_environment(),
                [f"analysis-{self.analysis_id}"],
                use_colour=False,
            )

        else:
            self.coloured_analysis_formatter = self.logger.handlers[0].formatter
            self.uncoloured_analysis_formatter = self.logger.handlers[0].formatter

    def __enter__(self):
        """Carry out the following:

        1. Remove the initial formatters from the logger's handlers
        2. Add the analysis formatter to the logger's current handlers and any extra handlers
        3. Add the extra handlers to the logger

        :return None:
        """
        if should_use_octue_log_handler():
            self._remove_formatters()
            self._add_analysis_formatter_to_log_handlers()

        self._add_extra_handlers()

    def __exit__(self, *args):
        """Carry out the following:

        1. Remove the extra handlers from the logger
        2. Remove the analysis log formatter from the logger's initial handlers
        3. Restore the initial formatters to the logger's initial handlers

        :return None:
        """
        self._remove_extra_handlers()

        if should_use_octue_log_handler():
            self._remove_formatters()
            self._restore_initial_formatters()

    def _remove_formatters(self):
        """Remove the formatters from any handlers the logger currently has.

        :return None:
        """
        for handler in self.logger.handlers:
            handler.formatter = None

    def _remove_extra_handlers(self):
        """Remove any extra handlers from the logger.

        :return None:
        """
        for handler in self.extra_log_handlers:
            self.logger.removeHandler(handler)

    def _add_analysis_formatter_to_log_handlers(self):
        """Apply the coloured analysis formatter to the logger's current handlers.

        :return None:
        """
        for handler in self.logger.handlers:
            handler.setFormatter(self.coloured_analysis_formatter)

    def _add_extra_handlers(self):
        """Add any extra log handlers to the logger and add the relevant analysis formatter to them.

        :return None:
        """
        if not self.extra_log_handlers:
            return

        # Apply any other given handlers to the logger.
        for extra_handler in self.extra_log_handlers:
            # Apply the uncoloured analysis formatter to any file log handlers to keep them readable (i.e. to avoid the
            # ANSI escape codes).
            if type(extra_handler).__name__ == "FileHandler":
                apply_log_handler(
                    logger=self.logger,
                    handler=extra_handler,
                    log_level=self.analysis_log_level,
                    formatter=self.uncoloured_analysis_formatter,
                )

            elif (
                type(extra_handler).__name__ == "MemoryHandler"
                and getattr(extra_handler, "target")
                and type(getattr(extra_handler, "target")).__name__ == "FileHandler"
            ):
                apply_log_handler(
                    logger=self.logger,
                    handler=extra_handler.target,
                    log_level=self.analysis_log_level,
                    formatter=self.uncoloured_analysis_formatter,
                )

            else:
                # Apply the coloured analysis formatter to any other types of log handler.
                apply_log_handler(
                    logger=self.logger,
                    handler=extra_handler,
                    log_level=self.analysis_log_level,
                    formatter=self.coloured_analysis_formatter,
                )

    def _restore_initial_formatters(self):
        """Restore the initial formatters to the logger's initial handlers.

        :return None:
        """
        for handler, formatter in self.initial_formatters:
            handler.setFormatter(formatter)



import copy
import json
import logging
import os

from octue.utils.decoders import OctueJSONDecoder
from octue.utils.encoders import OctueJSONEncoder


logger = logging.getLogger(__name__)


cached_local_metadata_files = {}
METADATA_FILENAME = ".octue"


class UpdateLocalMetadata:
    """A context manager that provides the contents of the given local metadata file and updates it with any changes
    made within its context. The local metadata is retrieved either from the disk or from the cache as appropriate.

    :param str path: the path to the local metadata. The file must be in JSON format.
    :return None:
    """

    def __init__(self, path=METADATA_FILENAME):
        self.path = path
        self._local_metadata = None

    def __enter__(self):
        """Load the local metadata file and return its contents.

        :return any: the contents of the local metadata file (converted from the JSON in the local metadata file)
        """
        self._local_metadata = load_local_metadata_file(self.path)
        return self._local_metadata

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Write any changes to the contents of the file.

        :return None:
        """
        overwrite_local_metadata_file(self._local_metadata, self.path)


def load_local_metadata_file(path=METADATA_FILENAME):
    """Load metadata from a local metadata records file, returning an empty dictionary if the file does not exist or is
    incorrectly formatted. If the file has already been cached, its contents are retrieved from the cache.

    :param str path: the path to the local metadata file
    :return dict: the contents of the local metadata file
    """
    absolute_path = os.path.abspath(path)
    cached_metadata = _get_metadata_from_cache(absolute_path)

    if cached_metadata:
        return cached_metadata

    if not os.path.exists(path):
        local_metadata = {}

    else:
        with open(path) as f:
            try:
                local_metadata = json.load(f, cls=OctueJSONDecoder)
            except json.decoder.JSONDecodeError:
                logger.warning(
                    f"The metadata file at {path!r} is incorrectly formatted so no metadata can be read from it. "
                    "Please fix or delete it."
                )
                local_metadata = {}

    _overwrite_cache_entry(absolute_path, local_metadata)
    return local_metadata


def overwrite_local_metadata_file(data, path=METADATA_FILENAME):
    """Create or overwrite the given local metadata file with the given data. If the data to overwrite the file with is
    the same as the file's cache entry, no changes are made. If it's not, the cache entry is updated and the file is
    overwritten.

    :param dict data: the data to overwrite the local metadata file with
    :param str path: the path to the local metadata file
    :return None:
    """
    absolute_path = os.path.abspath(path)
    cached_metadata = _get_metadata_from_cache(absolute_path)

    if data == cached_metadata:
        logger.debug("Avoiding overwriting local metadata file - its data is already in sync with the cache.")
        return

    _overwrite_cache_entry(absolute_path, data)

    with open(path, "w") as f:
        json.dump(data, f, cls=OctueJSONEncoder, indent=4)
        f.write("\n")


def _get_metadata_from_cache(absolute_path):
    """Get the metadata for the given local metadata file from the cache. If it's not cached, return `None`.

    :param str absolute_path: the path to the local metadata file
    :return dict|None: the metadata or, if the file hasn't been cached, `None`
    """
    logger.debug("Using cached local metadata.")
    return copy.deepcopy(cached_local_metadata_files.get(absolute_path))


def _overwrite_cache_entry(absolute_path, data):
    """Overwrite the metadata cache entry for the given local metadata file.

    :param str absolute_path: the path to the local metadata file
    :param dict data: the data to overwrite the existing cache entry with.
    :return None:
    """
    cached_local_metadata_files[absolute_path] = copy.deepcopy(data)
    logger.debug("Updated local metadata cache.")



import importlib
import logging
import os
import sys


logger = logging.getLogger(__name__)


class AppFrom:
    """A context manager that imports the module "app" from a file named "app.py" in the given directory on entry (by
    making a temporary addition to the system path) and unloads it (by deleting it from `sys.modules`) on exit. It will
    issue a warning if an existing module called "app" is already loaded. Usage example:

    ```python3
    with AppFrom('/path/to/dir') as app:
        Runner().run(app)
    ```

    :param str app_path: path to directory containing module named "app.py".
    :return None:
    """

    def __init__(self, app_path="."):
        self.app_path = os.path.abspath(os.path.normpath(app_path))
        logger.debug("Initialising AppFrom context at app_path %s", self.app_path)
        self.app_module = None

    def __enter__(self):
        # Warn on an app present on the system path
        if "app" in sys.modules.keys():
            logger.warning(
                "Module 'app' already on system path. Using 'AppFrom' context will yield unexpected results. Avoid "
                "using 'app' as a python module, except for your main entrypoint."
            )

        # Insert the present directory first on the system path.
        sys.path.insert(0, self.app_path)

        # Import the app from the present directory.
        self.app_module = importlib.import_module("app")

        # Immediately clean up the entry to the system path (don't use "remove" because if the user has it in their
        # path, this'll be an unexpected side effect, and don't do it in cleanup in case the called code inserts a path)
        sys.path.pop(0)
        logger.debug("Imported app at app_path and cleaned up temporary modification to sys.path %s", self.app_path)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Unload the imported module.

        :return None:
        """
        try:
            del sys.modules["app"]
            logger.debug("Deleted app from sys.modules")

        except KeyError:
            context_manager_name = type(self).__name__
            logger.warning(
                f"Module 'app' was already removed from the system path prior to exiting the {context_manager_name} "
                f"context manager. Using the {context_manager_name} context may yield unexpected results."
            )



def make_minimal_dictionary(**kwargs):
    """Make a dictionary with only the keyword arguments that have a non-`None` value.

    :param kwargs: any number of key-value pairs as keyword arguments
    :return dict: a dictionary containing only the key-value pairs that have a non-`None` value
    """
    data = {}

    for key, value in kwargs.items():
        if value is not None:
            data[key] = value

    return data



import abc
import importlib.metadata
import logging
import math
import os
import re
import time
from datetime import datetime

from octue.cloud import EXCEPTIONS_MAPPING
from octue.cloud.events.validation import SERVICE_COMMUNICATION_SCHEMA, is_event_valid
from octue.definitions import GOOGLE_COMPUTE_PROVIDERS
from octue.log_handlers import COLOUR_PALETTE
from octue.resources.manifest import Manifest


logger = logging.getLogger(__name__)


if os.environ.get("COMPUTE_PROVIDER", "UNKNOWN") in GOOGLE_COMPUTE_PROVIDERS:
    # Google Cloud logs don't support colour currently - provide a no-operation function.
    colourise = lambda string, text_colour=None, background_colour=None: string
else:
    from octue.utils.colour import colourise


PARENT_SDK_VERSION = importlib.metadata.version("octue")


class AbstractEventHandler:
    """An abstract event handler for Octue service events that:
    - Provide handlers for the Octue service event kinds (see https://strands.octue.com/octue/service-communication)
    - Handles received events in the order specified by the `order` attribute
    - Skips missing events after a set time and carries on handling from the next available event

    To create a concrete handler for a specific service/communication backend synchronously or asynchronously, inherit
    from this class and add the `handle_events` and `_extract_event_and_attributes` methods.

    :param octue.cloud.pub_sub.service.Service recipient: the `Service` instance that's receiving the events
    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param bool record_events: if `True`, record received events in the `received_events` attribute
    :param dict|None event_handlers: a mapping of event type names to callables that handle each type of event. The handlers must not mutate the events.
    :param dict schema: the JSON schema to validate events against
    :param int|float skip_missing_events_after: the number of seconds after which to skip any events if they haven't arrived but subsequent events have
    :param bool only_handle_result: if `True`, skip non-result events and only handle the "result" event when received
    :return None:
    """

    def __init__(
        self,
        recipient,
        handle_monitor_message=None,
        record_events=True,
        event_handlers=None,
        schema=SERVICE_COMMUNICATION_SCHEMA,
        skip_missing_events_after=10,
        only_handle_result=False,
    ):
        self.recipient = recipient
        self.handle_monitor_message = handle_monitor_message
        self.record_events = record_events
        self.schema = schema
        self.only_handle_result = only_handle_result

        # These are set when the first event is received.
        self.question_uuid = None
        self.child_sruid = None
        self.child_sdk_version = None

        self.waiting_events = None
        self.handled_events = []
        self._previous_event_number = -1
        self._start_time = None

        self.skip_missing_events_after = skip_missing_events_after
        self._missing_event_detection_time = None
        self._earliest_waiting_event_number = math.inf

        self._event_handlers = event_handlers or {
            "delivery_acknowledgement": self._handle_delivery_acknowledgement,
            "heartbeat": self._handle_heartbeat,
            "monitor_message": self._handle_monitor_message,
            "log_record": self._handle_log_message,
            "exception": self._handle_exception,
            "result": self._handle_result,
        }

        self._log_message_colours = [COLOUR_PALETTE[1], *COLOUR_PALETTE[3:]]

    @property
    def awaiting_missing_event(self):
        """Check if the event handler is currently waiting for a missing event.

        :return bool: `True` if the event handler is currently waiting for a missing event
        """
        return self._missing_event_detection_time is not None

    @property
    def time_since_missing_event(self):
        """Get the amount of time elapsed since the last missing event was detected. If no missing events have been
        detected or they've already been skipped, `None` is returned.

        :return float|None:
        """
        if not self.awaiting_missing_event:
            return None

        return time.perf_counter() - self._missing_event_detection_time

    @abc.abstractmethod
    def handle_events(self, *args, **kwargs):
        """Handle events and return a handled "result" event once one is received. This method must be overridden but
        can have any arguments. The first thing it should do is call `super().handle_events()`.

        :return dict: the handled final result
        """
        self.reset()

    def reset(self):
        """Reset the handler to be ready to handle a new stream of events.

        :return None:
        """
        self._start_time = time.perf_counter()
        self.waiting_events = {}
        self._previous_event_number = -1

    @abc.abstractmethod
    def _extract_event_and_attributes(self, container):
        """Extract an event and its attributes from the event container. This method must be overridden.

        :param any container: the container of the event (e.g. a Pub/Sub message)
        :return (any, dict): the event and its attributes (both must conform to the service communications event schema)
        """
        pass

    def _extract_and_enqueue_event(self, container):
        """Extract an event from its container, validate it, and add it to `self.waiting_events` if it's valid.

        :param any container: the container of the event (e.g. a Pub/Sub message)
        :return None:
        """
        try:
            event, attributes = self._extract_event_and_attributes(container)
        except Exception:
            event = None
            attributes = {}

        # Don't assume the presence of specific attributes before validation.
        child_sdk_version = attributes.get("sender_sdk_version")

        if not is_event_valid(
            event=event,
            attributes=attributes,
            recipient=self.recipient,
            parent_sdk_version=PARENT_SDK_VERSION,
            child_sdk_version=child_sdk_version,
            schema=self.schema,
        ):
            return

        # Get the child's SRUID and Octue SDK version from the first event.
        if not self.child_sdk_version:
            self.question_uuid = attributes["question_uuid"]
            self.child_sruid = attributes["sender"]
            self.child_sdk_version = attributes["sender_sdk_version"]

        logger.debug("%r: Received an event related to question %r.", self.recipient, self.question_uuid)
        order = attributes["order"]

        if order in self.waiting_events:
            logger.warning(
                "%r: Event with duplicate order %d received for question %r - overwriting original event.",
                self.recipient,
                order,
                self.question_uuid,
            )

        self.waiting_events[order] = event

    def _attempt_to_handle_waiting_events(self):
        """Attempt to handle any events waiting in `self.waiting_events`. If these events aren't consecutive to the
        last handled event (i.e. if events have been received out of order and the next in-order event hasn't been
        received yet), just return. After the missing event wait time has passed, if this set of missing events
        haven't arrived but subsequent ones have, skip to the earliest waiting event and continue from there.

        :return any|None: either a handled non-`None` "result" event, or `None` if nothing was returned by the event handlers or if the next in-order event hasn't been received yet
        """
        # Handle the case where no events (or no valid events) have been received.
        if not self.waiting_events:
            logger.debug("No events (or no valid events) were received.")
            return

        self._earliest_waiting_event_number = min(self.waiting_events.keys())

        while self.waiting_events:
            try:
                # If the next consecutive event has been received:
                event = self.waiting_events.pop(self._previous_event_number + 1)

            # If the next consecutive event hasn't been received:
            except KeyError:
                # Start the missing event timer if it isn't already running.
                if not self.awaiting_missing_event:
                    self._missing_event_detection_time = time.perf_counter()

                if self.time_since_missing_event > self.skip_missing_events_after:
                    event = self._skip_to_earliest_waiting_event()

                    # Declare there are no more missing events.
                    self._missing_event_detection_time = None

                    if not event:
                        return

                else:
                    return

            result = self._handle_event(event)

            if result is not None:
                return result

    def _skip_to_earliest_waiting_event(self):
        """Get the earliest waiting event and set the event handler up to continue from it.

        :return dict|None: the earliest waiting event if there is one
        """
        try:
            event = self.waiting_events.pop(self._earliest_waiting_event_number)
        except KeyError:
            return

        number_of_missing_events = self._earliest_waiting_event_number - self._previous_event_number - 1

        # Let the event handler know it can handle the next earliest event.
        self._previous_event_number = self._earliest_waiting_event_number - 1

        logger.warning(
            "%r: %d consecutive events missing for question %r after %ds - skipping to next earliest waiting event "
            "(event %d).",
            self.recipient,
            number_of_missing_events,
            self.question_uuid,
            self.skip_missing_events_after,
            self._earliest_waiting_event_number,
        )

        return event

    def _handle_event(self, event):
        """Pass an event to its handler and update the previous event number.

        :param dict event: the event to handle
        :return dict|None: the output of the event (this should be `None` unless the event is a "result" event)
        """
        self._previous_event_number += 1

        if self.record_events:
            self.handled_events.append(event)

        if self.only_handle_result and event["kind"] != "result":
            return

        handler = self._event_handlers[event["kind"]]
        return handler(event)

    def _handle_delivery_acknowledgement(self, event):
        """Log that the question was delivered.

        :param dict event:
        :return None:
        """
        logger.info("%r's question was delivered at %s.", self.recipient, event["datetime"])

    def _handle_heartbeat(self, event):
        """Record the time the heartbeat was received.

        :param dict event:
        :return None:
        """
        self._last_heartbeat = datetime.now()
        logger.info(
            "%r: Received a heartbeat from service %r for question %r.",
            self.recipient,
            self.child_sruid,
            self.question_uuid,
        )

    def _handle_monitor_message(self, event):
        """Send the monitor message to the handler if one has been provided.

        :param dict event:
        :return None:
        """
        logger.debug(
            "%r: Received a monitor message from service %r for question %r.",
            self.recipient,
            self.child_sruid,
            self.question_uuid,
        )

        if self.handle_monitor_message is not None:
            self.handle_monitor_message(event["data"])

    def _handle_log_message(self, event):
        """Deserialise the event into a log record and pass it to the local log handlers. The child's SRUID and the
        question UUID are added to the start of the log message, and the SRUIDs of any subchildren called by the child
        are each coloured differently.

        :param dict event:
        :return None:
        """
        record = logging.makeLogRecord(event["log_record"])

        # Add information about the immediate child sending the event and colour it with the first colour in the
        # colour palette.
        immediate_child_analysis_section = colourise(
            f"[{self.child_sruid} | analysis-{self.question_uuid}]",
            text_colour=self._log_message_colours[0],
        )

        # Colour any analysis sections from children of the immediate child with the rest of the colour palette.
        subchild_analysis_sections = [section.strip("[") for section in re.split("] ", record.msg)]
        final_message = subchild_analysis_sections.pop(-1)

        for i in range(len(subchild_analysis_sections)):
            subchild_analysis_sections[i] = colourise(
                "[" + subchild_analysis_sections[i] + "]",
                text_colour=self._log_message_colours[1:][i % len(self._log_message_colours[1:])],
            )

        record.msg = " ".join([immediate_child_analysis_section, *subchild_analysis_sections, final_message])
        logger.handle(record)

    def _handle_exception(self, event):
        """Raise the exception from the child.

        :param dict event:
        :raise Exception:
        :return None:
        """
        exception_message = "\n\n".join(
            (
                event["exception_message"],
                f"The following traceback was captured from the remote service {self.child_sruid!r}:",
                "".join(event["exception_traceback"]),
            )
        )

        try:
            exception_type = EXCEPTIONS_MAPPING[event["exception_type"]]

        # Allow unknown exception types to still be raised.
        except KeyError:
            exception_type = type(event["exception_type"], (Exception,), {})

        raise exception_type(exception_message)

    def _handle_result(self, event):
        """Extract any output values and output manifest from the result, deserialising the manifest if present.

        :param dict event:
        :return dict:
        """
        logger.info("%r: Received an answer to question %r.", self.recipient, self.question_uuid)

        if event.get("output_manifest"):
            output_manifest = Manifest.deserialise(event["output_manifest"])
        else:
            output_manifest = None

        return {"output_values": event.get("output_values"), "output_manifest": output_manifest}



import datetime
import logging
import os
import socket
from contextlib import closing

from gcp_storage_emulator.server import create_server
from google.cloud.storage._helpers import _get_default_storage_base_url


# Silence the GCP storage emulator logger below `ERROR` level messages.
emulator_logger = logging.getLogger("gcp_storage_emulator.server")
emulator_logger.setLevel(logging.ERROR)


class GoogleCloudStorageEmulator:
    """A local emulator for Google Cloud Storage

    :param str host:
    :param int port:
    :param str default_bucket:
    :return None:
    """

    def __init__(self, host="localhost", port=9090, in_memory=True, default_bucket=None):
        self._server = create_server(host, port, in_memory=in_memory, default_bucket=default_bucket)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def start(self):
        """Start the emulator. Do nothing if it's already started on the given host and port.

        :return None:
        """
        try:
            self._server.start()
        except RuntimeError:
            pass

    def stop(self):
        """Stop the emulator.

        :return None:
        """
        self._server.stop()


def get_free_tcp_port():
    """Get a free TCP port.

    :return int:
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        _, port = s.getsockname()
        return port


class GoogleCloudStorageEmulatorTestResultModifier:
    """A class providing `startTestRun` and `endTestRun` methods for use by a `unittest.TestResult` that start up a
    Google Cloud Storage emulator on a free port before the tests run and stop it after they've all run.

    :param str host:
    :param bool in_memory:
    :param str default_bucket_name:
    :return None:
    """

    STORAGE_EMULATOR_HOST_ENVIRONMENT_VARIABLE_NAME = "STORAGE_EMULATOR_HOST"

    def __init__(self, host="localhost", in_memory=True, default_bucket_name=None):
        port = get_free_tcp_port()
        self.storage_emulator_host = f"http://{host}:{port}"

        self.storage_emulator = GoogleCloudStorageEmulator(
            host=host, port=port, in_memory=in_memory, default_bucket=default_bucket_name
        )

    def startTestRun(self):
        """Start the Google Cloud Storage emulator before starting the test run.

        :param unittest.TestResult test_result:
        :return None:
        """
        os.environ[self.STORAGE_EMULATOR_HOST_ENVIRONMENT_VARIABLE_NAME] = self.storage_emulator_host
        self.storage_emulator.start()

    def stopTestRun(self):
        """Stop the Google Cloud Storage emulator before starting the test run.

        :param unittest.TestResult test_result:
        :return None:
        """
        self.storage_emulator.stop()
        del os.environ[self.STORAGE_EMULATOR_HOST_ENVIRONMENT_VARIABLE_NAME]


def mock_generate_signed_url(
    blob,
    expiration=datetime.timedelta(days=7),
    api_access_endpoint=_get_default_storage_base_url(),
    **kwargs,
):
    """Mock generating a signed URL for a Google Cloud Storage blob. Signed URLs can't currently be generated when using
    workload identity federation, which we use for our CI tests.

    :param google.cloud.storage.blob.Blob blob:
    :param datetime.datetime|datetime.timedelta expiration:
    :param str api_access_endpoint:
    :return str:
    """
    mock_signed_query_parameter = (
        f"?Expires={round((datetime.datetime.now() + expiration).timestamp())}&GoogleAccessId=my-service-account%40my-p"
        f"roject.iam.gserviceaccount.com&Signature=mock-signature"
    )

    base_url = "/".join((api_access_endpoint, blob.bucket.name, blob.name))
    return base_url + mock_signed_query_parameter



import logging

from octue.cloud.events.handler import AbstractEventHandler
from octue.cloud.events.validation import SERVICE_COMMUNICATION_SCHEMA
from octue.cloud.pub_sub.service import Service
from octue.resources.service_backends import ServiceBackend


logger = logging.getLogger(__name__)


class EventReplayer(AbstractEventHandler):
    """A replayer for events retrieved asynchronously from storage. Missing events are immediately skipped.

    :param octue.cloud.pub_sub.service.Service recipient: the `Service` instance that's receiving the events
    :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive
    :param bool record_events: if `True`, record received events in the `received_events` attribute
    :param dict|None event_handlers: a mapping of event type names to callables that handle each type of event. The handlers must not mutate the events.
    :param dict|str schema: the JSON schema to validate events against
    :param bool only_handle_result: if `True`, skip non-result events and only handle the "result" event if present
    :return None:
    """

    def __init__(
        self,
        recipient=None,
        handle_monitor_message=None,
        record_events=True,
        event_handlers=None,
        schema=SERVICE_COMMUNICATION_SCHEMA,
        only_handle_result=False,
    ):
        super().__init__(
            recipient or Service(backend=ServiceBackend(), service_id="local/local:local"),
            handle_monitor_message=handle_monitor_message,
            record_events=record_events,
            event_handlers=event_handlers,
            schema=schema,
            skip_missing_events_after=0,
            only_handle_result=only_handle_result,
        )

    def handle_events(self, events):
        """Handle the given events and return a handled "result" event if one is present.

        :param iter(dict) events: the events to handle
        :return dict|None: the handled "result" event if present
        """
        super().handle_events()

        for event in events:
            self._extract_and_enqueue_event(event)

        return self._attempt_to_handle_waiting_events()

    def _extract_event_and_attributes(self, container):
        """Extract an event and its attributes from the event container.

        :param dict container: the container of the event
        :return (any, dict): the event and its attributes
        """
        container["attributes"]["order"] = int(container["attributes"]["order"])
        return container["event"], container["attributes"]



import json

from google.cloud.bigquery import Client, QueryJobConfig, ScalarQueryParameter

from octue.cloud.events.validation import VALID_EVENT_KINDS


def get_events(
    table_id,
    sender,
    question_uuid,
    kind=None,
    include_attributes=False,
    include_backend_metadata=False,
    limit=1000,
):
    """Get Octue service events for a question from a sender from a Google BigQuery event store.

    :param str table_id: the full ID of the table e.g. "your-project.your-dataset.your-table"
    :param str sender: the SRUID of the sender of the events
    :param str question_uuid: the UUID of the question to get the events for
    :param str|None kind: the kind of event to get; if `None`, all event kinds are returned
    :param bool include_attributes: if `True`, include events' attributes (excluding question UUID)
    :param bool include_backend_metadata: if `True`, include the service backend metadata
    :param int limit: the maximum number of events to return
    :return list(dict): the events for the question
    """
    if kind:
        if kind not in VALID_EVENT_KINDS:
            raise ValueError(f"`kind` must be one of {VALID_EVENT_KINDS!r}; received {kind!r}.")

        event_kind_condition = [f'AND JSON_EXTRACT_SCALAR(event, "$.kind") = "{kind}"']
    else:
        event_kind_condition = []

    client = Client()
    fields = ["`event`"]

    if include_attributes:
        fields.extend(
            (
                "`datetime`",
                "`uuid`",
                "`originator`",
                "`sender`",
                "`sender_type`",
                "`sender_sdk_version`",
                "`recipient`",
                "`order`",
                "`other_attributes`",
            )
        )

    if include_backend_metadata:
        fields.extend(("`backend`", "`backend_metadata`"))

    query = "\n".join(
        [
            f"SELECT {', '.join(fields)} FROM `{table_id}`",
            "WHERE sender=@sender",
            "AND question_uuid=@question_uuid",
            *event_kind_condition,
            "ORDER BY `order`",
            "LIMIT @limit",
        ]
    )

    job_config = QueryJobConfig(
        query_parameters=[
            ScalarQueryParameter("sender", "STRING", sender),
            ScalarQueryParameter("question_uuid", "STRING", question_uuid),
            ScalarQueryParameter("limit", "INTEGER", limit),
        ]
    )

    query_job = client.query(query, job_config=job_config)
    rows = query_job.result()
    df = rows.to_dataframe()

    # Convert JSON strings to python primitives.
    df["event"] = df["event"].map(json.loads)

    if "other_attributes" in df:
        df["other_attributes"] = df["other_attributes"].map(json.loads)

    if "backend_metadata" in df:
        df["backend_metadata"] = df["backend_metadata"].map(json.loads)

    return df.to_dict(orient="records")



import concurrent.futures
import copy
import datetime
import functools
import importlib.metadata
import json
import logging
import threading
import uuid

import google.api_core.exceptions
import jsonschema
from google.api_core import retry
from google.cloud import pubsub_v1

import octue.exceptions
from octue.cloud.events import OCTUE_SERVICES_PREFIX
from octue.cloud.events.counter import EventCounter
from octue.cloud.events.validation import raise_if_event_is_invalid
from octue.cloud.pub_sub import Subscription, Topic
from octue.cloud.pub_sub.events import GoogleCloudPubSubEventHandler, extract_event_and_attributes_from_pub_sub_message
from octue.cloud.pub_sub.logging import GoogleCloudPubSubHandler
from octue.cloud.service_id import (
    convert_service_id_to_pub_sub_form,
    create_sruid,
    get_default_sruid,
    raise_if_revision_not_registered,
    split_service_id,
    validate_sruid,
)
from octue.compatibility import warn_if_incompatible
from octue.utils.dictionaries import make_minimal_dictionary
from octue.utils.encoders import OctueJSONEncoder
from octue.utils.exceptions import convert_exception_to_primitives
from octue.utils.threads import RepeatingTimer


logger = logging.getLogger(__name__)

# A lock to ensure only one event can be emitted at a time so that the order is incremented correctly when events are
# being emitted on multiple threads (e.g. via the main thread and a periodic monitor message thread). This avoids 1)
# events overwriting each other in the parent's message handler and 2) events losing their order.
emit_event_lock = threading.Lock()

DEFAULT_NAMESPACE = "default"
ANSWERS_NAMESPACE = "answers"
OCTUE_SERVICE_REGISTRY_ENDPOINT = "services.registry.octue.com"

# Switch message batching off by setting `max_messages` to 1. This minimises latency and is recommended for
# microservices publishing single messages in a request-response sequence.
BATCH_SETTINGS = pubsub_v1.types.BatchSettings(max_bytes=10 * 1000 * 1000, max_latency=0.01, max_messages=1)

PARENT_SENDER_TYPE = "PARENT"
CHILD_SENDER_TYPE = "CHILD"


class Service:
    """An Octue service that can be used as a data service or digital twin in one of two modes:

    - As a child accepting questions (input values and manifests) from parents, running them through its app, and
      responding with an answer
    - As a parent asking questions to children in the above mode

    Services communicate entirely via Google Pub/Sub and can ask and/or respond to questions from any other service that
    has a corresponding topic on Google Pub/Sub.

    :param octue.resources.service_backends.ServiceBackend backend: the object representing the type of backend the service uses
    :param str|None service_id: a unique ID to give to the service (any string); a UUID is generated if none is given
    :param callable|None run_function: the function the service should run when it is called
    :param str|None name: an optional name to use for the service to override its ID in its string representation
    :param iter(dict)|None service_registries: the names and endpoints of the registries used to resolve service revisions when asking questions; these should be in priority order (highest priority first)
    :return None:
    """

    def __init__(self, backend, service_id=None, run_function=None, name=None, service_registries=None):
        if service_id is None:
            self.id = create_sruid(namespace=DEFAULT_NAMESPACE, name=str(uuid.uuid4()))

        # Raise an error if the service ID is some kind of falsey object that isn't `None`.
        elif not service_id:
            raise ValueError(f"`service_id` should be `None` or a non-falsey value; received {service_id!r} instead.")

        else:
            validate_sruid(service_id)
            self.id = service_id

        self.backend = backend
        self.run_function = run_function
        self.name = name
        self.service_registries = service_registries

        self._pub_sub_id = convert_service_id_to_pub_sub_form(self.id)
        self._local_sdk_version = importlib.metadata.version("octue")
        self._publisher = None
        self._services_topic = None
        self._event_handler = None

    def __repr__(self):
        """Represent the service as a string.

        :return str: the service represented as a string
        """
        return f"<{type(self).__name__}({self.name or self.id!r})>"

    @property
    def publisher(self):
        """Get or instantiate the publisher client for the service. No publisher is instantiated until this property is
        called for the first time. This allows checking for the `GOOGLE_APPLICATION_CREDENTIALS` environment variable to
        be put off until it's needed.

        :return google.cloud.pubsub_v1.PublisherClient:
        """
        if not self._publisher:
            self._publisher = pubsub_v1.PublisherClient(batch_settings=BATCH_SETTINGS)

        return self._publisher

    @property
    def services_topic(self):
        """Get the Octue services topic that all events in the project are published to. No topic is instantiated until
        this property is called for the first time. This allows checking for the `GOOGLE_APPLICATION_CREDENTIALS`
        environment variable to be put off until it's needed.

        :raise octue.exceptions.ServiceNotFound: if the topic doesn't exist in the project
        :return octue.cloud.pub_sub.topic.Topic: the Octue services topic for the project
        """
        if not self._services_topic:
            topic = Topic(name=OCTUE_SERVICES_PREFIX, project_name=self.backend.project_name)

            if not topic.exists():
                raise octue.exceptions.ServiceNotFound(f"{topic!r} cannot be found.")

            self._services_topic = topic

        return self._services_topic

    @property
    def received_events(self):
        """Get the events received from a child service while running the `wait_for_answer` method. If the
        `wait_for_answer` method hasn't been run, `None` is returned. If an empty list is returned, no events have been
        received.

        :return list(dict)|None:
        """
        if self._event_handler:
            return self._event_handler.handled_events
        return None

    def serve(self, timeout=None, delete_topic_and_subscription_on_exit=False, allow_existing=False, detach=False):
        """Start the service as a child, waiting to accept questions from any other Octue service using Google Pub/Sub
        on the same Google Cloud project. Questions are accepted, processed, and answered asynchronously.

        :param float|None timeout: time in seconds after which to shut down the child
        :param bool delete_topic_and_subscription_on_exit: if `True`, delete the child's topic and subscription on exiting serving mode
        :param bool allow_existing: if `True`, allow starting a service for which the topic and/or subscription already exists (indicating an existing service) - this connects this service to the existing service's topic and subscription
        :param bool detach: if `True`, detach from the subscription future. The future and subscriber are returned so can still be cancelled and closed manually. Note that the topic and subscription are not automatically deleted on exit if this option is chosen.
        :return (google.cloud.pubsub_v1.subscriber.futures.StreamingPullFuture, google.cloud.pubsub_v1.SubscriberClient):
        """
        logger.info("Starting %r.", self)

        subscription = Subscription(
            name=".".join((OCTUE_SERVICES_PREFIX, self._pub_sub_id)),
            topic=self.services_topic,
            filter=f'attributes.recipient = "{self.id}" AND attributes.sender_type = "{PARENT_SENDER_TYPE}"',
            expiration_time=None,
        )

        try:
            subscription.create(allow_existing=allow_existing)
        except google.api_core.exceptions.AlreadyExists:
            raise octue.exceptions.ServiceAlreadyExists(f"A service with the ID {self.id!r} already exists.")

        subscriber = pubsub_v1.SubscriberClient()

        try:
            future = subscriber.subscribe(subscription=subscription.path, callback=self.answer)

            logger.info(
                "You can now ask this service questions at %r using the `octue.resources.Child` class.",
                self.id,
            )

            # If not detaching, keep answering questions until the subscriber times out (or forever if there's no
            # timeout).
            if not detach:
                try:
                    future.result(timeout=timeout)
                except (TimeoutError, concurrent.futures.TimeoutError, KeyboardInterrupt):
                    future.cancel()

        finally:
            # If not detaching, delete the topic and subscription deletion if required and close the subscriber.
            if not detach:
                if delete_topic_and_subscription_on_exit:
                    try:
                        if subscription.creation_triggered_locally:
                            subscription.delete()

                    except Exception:
                        logger.error("Deletion of %r failed.", subscription)

                subscriber.close()

        return future, subscriber

    def answer(self, question, heartbeat_interval=120, timeout=30):
        """Answer a question from a parent - i.e. run the child's app on the given data and return the output values.
        Answers conform to the output values and output manifest schemas specified in the child's Twine file.

        :param dict|google.cloud.pubsub_v1.subscriber.message.Message question:
        :param int|float heartbeat_interval: the time interval, in seconds, at which to send heartbeats
        :param float|None timeout: time in seconds to keep retrying sending of the answer once it has been calculated
        :raise Exception: if any exception arises during running analysis and sending its results
        :return None:
        """
        try:
            (
                question,
                question_uuid,
                forward_logs,
                parent_sdk_version,
                save_diagnostics,
                originator,
            ) = self._parse_question(question)
        except jsonschema.ValidationError:
            return

        heartbeater = None
        order = EventCounter()

        try:
            self._send_delivery_acknowledgment(question_uuid, originator, order)

            heartbeater = RepeatingTimer(
                interval=heartbeat_interval,
                function=self._send_heartbeat,
                kwargs={"question_uuid": question_uuid, "originator": originator, "order": order},
            )

            heartbeater.daemon = True
            heartbeater.start()

            if forward_logs:
                analysis_log_handler = GoogleCloudPubSubHandler(
                    event_emitter=self._emit_event,
                    question_uuid=question_uuid,
                    originator=originator,
                    recipient=originator,
                    order=order,
                )
            else:
                analysis_log_handler = None

            analysis = self.run_function(
                analysis_id=question_uuid,
                input_values=question.get("input_values"),
                input_manifest=question.get("input_manifest"),
                children=question.get("children"),
                analysis_log_handler=analysis_log_handler,
                handle_monitor_message=functools.partial(
                    self._send_monitor_message,
                    question_uuid=question_uuid,
                    originator=originator,
                    order=order,
                ),
                save_diagnostics=save_diagnostics,
            )

            result = make_minimal_dictionary(kind="result", output_values=analysis.output_values)

            if analysis.output_manifest is not None:
                result["output_manifest"] = analysis.output_manifest.to_primitive()

            self._emit_event(
                event=result,
                originator=originator,
                recipient=originator,
                order=order,
                attributes={"question_uuid": question_uuid, "sender_type": CHILD_SENDER_TYPE},
                timeout=timeout,
            )

            heartbeater.cancel()
            logger.info("%r answered question %r.", self, question_uuid)

        except BaseException as error:  # noqa
            if heartbeater is not None:
                heartbeater.cancel()

            warn_if_incompatible(child_sdk_version=self._local_sdk_version, parent_sdk_version=parent_sdk_version)
            self.send_exception(question_uuid, originator, order, timeout=timeout)
            raise error

    def ask(
        self,
        service_id,
        input_values=None,
        input_manifest=None,
        children=None,
        subscribe_to_logs=True,
        allow_local_files=False,
        save_diagnostics="SAVE_DIAGNOSTICS_ON_CRASH",  # This is repeated as a string here to avoid a circular import.
        question_uuid=None,
        push_endpoint=None,
        asynchronous=False,
        timeout=86400,
    ):
        """Ask a child a question (i.e. send it input values for it to analyse and produce output values for) and return
        a subscription to receive its answer on. The input values and manifest must conform to the schemas in the
        child's Twine file.

        :param str service_id: the ID of the child to ask the question to
        :param any|None input_values: any input values for the question
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys.
        :param bool subscribe_to_logs: if `True`, subscribe to the child's logs and handle them with the local log handlers
        :param bool allow_local_files: if `True`, allow the input manifest to contain references to local files - this should only be set to `True` if the child will be able to access these local files
        :param str save_diagnostics: must be one of {"SAVE_DIAGNOSTICS_OFF", "SAVE_DIAGNOSTICS_ON_CRASH", "SAVE_DIAGNOSTICS_ON"}; if turned on, allow the input values and manifest (and its datasets) to be saved by the child either all the time or just if it fails while processing them
        :param str|None question_uuid: the UUID to use for the question if a specific one is needed; a UUID is generated if not
        :param str|None push_endpoint: if answers to the question should be pushed to an endpoint, provide its URL here (the returned subscription will be a push subscription); if not, leave this as `None`
        :param bool asynchronous: if `True` and not using a push endpoint, don't create an answer subscription
        :param float|None timeout: time in seconds to keep retrying sending the question
        :return (octue.cloud.pub_sub.subscription.Subscription|None, str): the answer subscription (if the question is synchronous or a push endpoint was used) and question UUID
        """
        service_namespace, service_name, service_revision_tag = split_service_id(service_id)

        if self.service_registries:
            if service_revision_tag:
                raise_if_revision_not_registered(sruid=service_id, service_registries=self.service_registries)
            else:
                service_id = get_default_sruid(
                    namespace=service_namespace,
                    name=service_name,
                    service_registries=self.service_registries,
                )

        elif not service_revision_tag:
            raise octue.exceptions.InvalidServiceID(
                f"A service revision tag for {service_id!r} must be provided if service registries aren't being used."
            )

        if not allow_local_files:
            if (input_manifest is not None) and (not input_manifest.all_datasets_are_in_cloud):
                raise octue.exceptions.FileLocationError(
                    "All datasets of the input manifest and all files of the datasets must be uploaded to the cloud "
                    "before asking a service to perform an analysis upon them. The manifest must then be updated with "
                    "the new cloud locations."
                )

        question_uuid = question_uuid or str(uuid.uuid4())

        if asynchronous and not push_endpoint:
            answer_subscription = None
        else:
            pub_sub_id = convert_service_id_to_pub_sub_form(self.id)

            answer_subscription = Subscription(
                name=".".join((OCTUE_SERVICES_PREFIX, pub_sub_id, ANSWERS_NAMESPACE, question_uuid)),
                topic=self.services_topic,
                filter=(
                    f'attributes.recipient = "{self.id}" '
                    f'AND attributes.question_uuid = "{question_uuid}" '
                    f'AND attributes.sender_type = "{CHILD_SENDER_TYPE}"'
                ),
                push_endpoint=push_endpoint,
            )
            answer_subscription.create(allow_existing=False)

        self._send_question(
            input_values=input_values,
            input_manifest=input_manifest,
            children=children,
            forward_logs=subscribe_to_logs,
            save_diagnostics=save_diagnostics,
            question_uuid=question_uuid,
            recipient=service_id,
        )

        return answer_subscription, question_uuid

    def wait_for_answer(
        self,
        subscription,
        handle_monitor_message=None,
        record_events=True,
        timeout=60,
        maximum_heartbeat_interval=300,
    ):
        """Wait for an answer to a question on the given subscription, deleting the subscription and its topic once
        the answer is received.

        :param octue.cloud.pub_sub.subscription.Subscription subscription: the subscription for the question's answer
        :param callable|None handle_monitor_message: a function to handle monitor messages (e.g. send them to an endpoint for plotting or displaying) - this function should take a single JSON-compatible python primitive as an argument (note that this could be an array or object)
        :param bool record_events: if `True`, record messages received from the child in the `received_events` attribute
        :param float|None timeout: how long in seconds to wait for an answer before raising a `TimeoutError`
        :param float|int delivery_acknowledgement_timeout: how long in seconds to wait for a delivery acknowledgement before aborting
        :param float|int maximum_heartbeat_interval: the maximum amount of time (in seconds) allowed between child heartbeats before an error is raised
        :raise TimeoutError: if the timeout is exceeded
        :return dict: dictionary containing the keys "output_values" and "output_manifest"
        """
        if subscription.is_push_subscription:
            raise octue.exceptions.NotAPullSubscription(
                f"{subscription.path!r} is a push subscription so it cannot be waited on for an answer. Please check "
                f"its push endpoint at {subscription.push_endpoint!r}."
            )

        self._event_handler = GoogleCloudPubSubEventHandler(
            subscription=subscription,
            recipient=self,
            handle_monitor_message=handle_monitor_message,
            record_events=record_events,
        )

        try:
            return self._event_handler.handle_events(
                timeout=timeout,
                maximum_heartbeat_interval=maximum_heartbeat_interval,
            )

        finally:
            subscription.delete()

    def send_exception(self, question_uuid, originator, order, timeout=30):
        """Serialise and send the exception being handled to the parent.

        :param str question_uuid: the UUID of the question this event relates to
        :param str originator: the SRUID of the service that asked the question this event is related to
        :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
        :param float|None timeout: time in seconds to keep retrying sending of the exception
        :return None:
        """
        exception = convert_exception_to_primitives()
        exception_message = f"Error in {self!r}: {exception['message']}"

        self._emit_event(
            {
                "kind": "exception",
                "exception_type": exception["type"],
                "exception_message": exception_message,
                "exception_traceback": exception["traceback"],
            },
            originator=originator,
            recipient=originator,
            order=order,
            attributes={"question_uuid": question_uuid, "sender_type": CHILD_SENDER_TYPE},
            timeout=timeout,
        )

    def _emit_event(self, event, originator, recipient, order, attributes=None, timeout=30):
        """Emit a JSON-serialised event as a Pub/Sub message to the services topic with optional message attributes,
        incrementing the `order` argument by one. This method is thread-safe.

        :param dict event: JSON-serialisable data to emit as an event
        :param str originator: the SRUID of the service that asked the question this event is related to
        :param str recipient: the SRUID of the service the event is intended for
        :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
        :param dict|None attributes: key-value pairs to attach to the event - the values must be strings or bytes
        :param int|float timeout: the timeout for sending the event in seconds
        :return google.cloud.pubsub_v1.publisher.futures.Future:
        """
        attributes = attributes or {}
        attributes["uuid"] = str(uuid.uuid4())
        attributes["originator"] = originator
        attributes["sender"] = self.id
        attributes["sender_sdk_version"] = self._local_sdk_version
        attributes["recipient"] = recipient

        with emit_event_lock:
            attributes["order"] = int(order)
            attributes["datetime"] = datetime.datetime.utcnow().isoformat()
            converted_attributes = {}

            for key, value in attributes.items():
                if isinstance(value, bool):
                    value = str(int(value))
                elif isinstance(value, (int, float)):
                    value = str(value)

                converted_attributes[key] = value

            future = self.publisher.publish(
                topic=self.services_topic.path,
                data=json.dumps(event, cls=OctueJSONEncoder).encode(),
                retry=retry.Retry(deadline=timeout),
                **converted_attributes,
            )

            order += 1

        return future

    def _send_question(
        self,
        input_values,
        input_manifest,
        children,
        forward_logs,
        save_diagnostics,
        question_uuid,
        recipient,
        timeout=30,
    ):
        """Send a question to a child service.

        :param any|None input_values: any input values for the question
        :param octue.resources.manifest.Manifest|None input_manifest: an input manifest of any datasets needed for the question
        :param list(dict)|None children: a list of children for the child to use instead of its default children (if it uses children). These should be in the same format as in an app's app configuration file and have the same keys.
        :param bool forward_logs: whether to request the child to forward its logs
        :param str save_diagnostics: must be one of {"SAVE_DIAGNOSTICS_OFF", "SAVE_DIAGNOSTICS_ON_CRASH", "SAVE_DIAGNOSTICS_ON"}; if turned on, allow the input values and manifest (and its datasets) to be saved by the child either all the time or just if it fails while processing them
        :param str question_uuid: the UUID of the question being sent
        :param str recipient: the SRUID of the child the question is intended for
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        question = make_minimal_dictionary(kind="question", input_values=input_values, children=children)

        if input_manifest is not None:
            input_manifest.use_signed_urls_for_datasets()
            question["input_manifest"] = input_manifest.to_primitive()

        future = self._emit_event(
            event=question,
            timeout=timeout,
            originator=self.id,
            recipient=recipient,
            order=EventCounter(),
            attributes={
                "question_uuid": question_uuid,
                "forward_logs": forward_logs,
                "save_diagnostics": save_diagnostics,
                "sender_type": PARENT_SENDER_TYPE,
            },
        )

        # Await successful publishing of the question.
        future.result()
        logger.info("%r asked a question %r to service %r.", self, question_uuid, recipient)

    def _send_delivery_acknowledgment(self, question_uuid, originator, order, timeout=30):
        """Send an acknowledgement of question receipt to the parent.

        :param str question_uuid: the UUID of the question this event relates to
        :param str originator: the SRUID of the service that asked the question this event is related to
        :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        self._emit_event(
            {
                "kind": "delivery_acknowledgement",
                "datetime": datetime.datetime.utcnow().isoformat(),
            },
            timeout=timeout,
            originator=originator,
            recipient=originator,
            order=order,
            attributes={"question_uuid": question_uuid, "sender_type": CHILD_SENDER_TYPE},
        )

        logger.info("%r acknowledged receipt of question %r.", self, question_uuid)

    def _send_heartbeat(self, question_uuid, originator, order, timeout=30):
        """Send a heartbeat to the parent, indicating that the service is alive.

        :param str question_uuid: the UUID of the question this event relates to
        :param str originator: the SRUID of the service that asked the question this event is related to
        :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
        :param float timeout: time in seconds after which to give up sending
        :return None:
        """
        self._emit_event(
            {
                "kind": "heartbeat",
                "datetime": datetime.datetime.utcnow().isoformat(),
            },
            originator=originator,
            recipient=originator,
            order=order,
            timeout=timeout,
            attributes={"question_uuid": question_uuid, "sender_type": CHILD_SENDER_TYPE},
        )

        logger.debug("Heartbeat sent by %r.", self)

    def _send_monitor_message(self, data, question_uuid, originator, order, timeout=30):
        """Send a monitor message to the parent.

        :param any data: the data to send as a monitor message
        :param str question_uuid: the UUID of the question this event relates to
        :param str originator: the SRUID of the service that asked the question this event is related to
        :param octue.cloud.events.counter.EventCounter order: an event counter keeping track of the order of emitted events
        :param float timeout: time in seconds to retry sending the message
        :return None:
        """
        self._emit_event(
            {"kind": "monitor_message", "data": data},
            originator=originator,
            recipient=originator,
            order=order,
            timeout=timeout,
            attributes={"question_uuid": question_uuid, "sender_type": CHILD_SENDER_TYPE},
        )

        logger.debug("Monitor message sent by %r.", self)

    def _parse_question(self, question):
        """Parse a question in the Google Cloud Run or Google Pub/Sub format.

        :param dict|google.cloud.pubsub_v1.subscriber.message.Message question: the question to parse in Google Cloud Run or Google Pub/Sub format
        :return (dict, str, bool, str, str, str): the question's event and its attributes (question UUID, whether to forward logs, the Octue SDK version of the parent, whether to save diagnostics, and the SRUID of the service revision that asked the question)
        """
        logger.info("%r received a question.", self)

        # Acknowledge the question if it's directly from Pub/Sub.
        if hasattr(question, "ack"):
            question.ack()

        event, attributes = extract_event_and_attributes_from_pub_sub_message(question)
        event_for_validation = copy.deepcopy(event)

        raise_if_event_is_invalid(
            event=event_for_validation,
            attributes=attributes,
            recipient=self,
            # Don't assume the presence of specific attributes before validation.
            parent_sdk_version=attributes.get("sender_sdk_version"),
            child_sdk_version=importlib.metadata.version("octue"),
        )

        logger.info("%r parsed question %r successfully.", self, attributes["question_uuid"])

        return (
            event,
            attributes["question_uuid"],
            attributes["forward_logs"],
            attributes["sender_sdk_version"],
            attributes["save_diagnostics"],
            attributes["originator"],
        )
